"""
The aurori project

Copyright (C) 2022  Marcus Drobisch,

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

__authors__ = ["Marcus Drobisch"]
__contact__ = "aurori@fabba.space"
__credits__ = []
__license__ = "AGPLv3+"

import logging
from datetime import datetime
from apscheduler.schedulers.asyncio  import AsyncIOScheduler
from cron_descriptor import ExpressionDescriptor, Options, CasingTypeEnum

from aurori.logs import logManager
from aurori.types.objDict import ObjDict
from aurori.database import db
from pytz import utc

class JobManager(object):
    """ The JobManager ...
    """
    def __init__(self, ):
        # preparation to instanciate
        self.config = None
        self.workspaceManager = None
        self.job_counter = 0
        self.jobs = {}
        self.scheduler = AsyncIOScheduler()
        logging.getLogger('apscheduler.scheduler').setLevel(logging.ERROR)
        logging.getLogger('apscheduler.scheduler').name = "SCHEDULER"
        logging.getLogger('apscheduler.executors.default').setLevel(logging.ERROR)
        logging.getLogger('apscheduler.executors.default').name = "EXECUTOR"
        logging.getLogger('EXECUTOR').setLevel(logging.ERROR)

    def init_manager(self, config):
        self.config = config

        from aurori.jobs.models import JobExecute
        self.job = JobExecute

    def get_jobs(self):
        return self.jobs

    def register_job(self, workspace, job_class, log_in_db=False):
        jobkey = job_class.__module__

        jobInstance = job_class()

        jobInstance.job_key = jobkey
        if workspace is not None:
            jobInstance.workspace = workspace.name
        job = {
            'job_class': job_class,
            'name': jobInstance.name,
            'workspace': workspace.name,
            'description': jobInstance.description,
            'parameters': jobInstance.parameters,
            'trigger': 'Internal',
            'log_in_db': log_in_db,
            'cron': jobInstance.cron,
            'day': jobInstance.day,
            'week': jobInstance.week,
            'day_of_week': jobInstance.day_of_week,
            'hour': jobInstance.hour,
            'minute': jobInstance.minute,
            'second': jobInstance.second,
        }

        if jobInstance.cron is True:
            cron_list = []

            if job_class.minute is None:
                cron_list.append("*")
            else:
                cron_list.append(job_class.minute)

            if job_class.hour is None:
                cron_list.append("*")
            else:
                cron_list.append(job_class.hour)

            if job_class.day is None:
                cron_list.append("*")
            else:
                cron_list.append(job_class.day)

            cron_list.append("*")

            if job_class.day_of_week is None:
                cron_list.append("*")
            else:
                cron_list.append(job_class.day_of_week)

            cron_string = " ".join(cron_list)
            options = Options()
            options.throw_exception_on_parse_error = False
            options.day_of_week_start_index_zero = True
            options.use_24hour_time_format = True
            options.casing_type = CasingTypeEnum.LowerCase
            descripter = ExpressionDescriptor(cron_string, options)
            logManager.info("Register repetitive job '{}' triggered {}".format(
                jobkey, descripter.get_description()))
            self.scheduler.add_job(
                jobInstance.start_job,
                kwargs=({
                    "job_id": str(jobkey)
                }),
                id=(str(jobkey)),
                trigger='cron',
                replace_existing=True,
                day=job_class.day,
                day_of_week=job_class.day_of_week,
                week=job_class.week,
                hour=job_class.hour,
                minute=job_class.minute,
                second=job_class.second,
            )
            job['trigger'] = descripter.get_description()

        self.jobs[str(jobkey)] = ObjDict(job.copy())

    def run_job(self,
                user,
                job_class,
                args,
                date,
                max_instances=10,
                log_trigger=False):
        job_key = job_class.__module__
        if job_key in self.jobs:
            with db.get_session() as db_session:
           
                je = None
                if log_trigger is True:
                    from aurori.jobs.models import JobExecute
                    je = JobExecute()
                    je.triggered_on = datetime.now()
                    if user is None:
                        je.triggered_by = ""
                    else:
                        je.triggered_by = user.email
                    je.name = job_key
                    je.workspace = self.jobs[job_key].workspace
                    je.state = "TRIGGERED"

                    db_session.add(je)
                    db_session.commit()

                # if self.jobs[jobkey]['cron']:
                #     # handle a cron job
                #     job = self.scheduler.get_job(jobkey)
                #     job.modify(next_run_time=datetime.now())
                #     return None
                # else:

                # handle a single trigger job
                jobInstance = self.jobs[job_key]['job_class']()
                self.job_counter += 1
                job_ececution_id = None

                if je is not None:
                    job_ececution_id = je.id

                kwargs = {
                    "job_id": str(job_key) + str(self.job_counter),
                    "job_execution_id": job_ececution_id
                }
                kwargs = {**kwargs, **args}
                self.scheduler.add_job(
                    jobInstance.start_job,
                    id=(str(job_key) + str(self.job_counter)),
                    trigger='date',
                    next_run_time=date.astimezone(utc),
                    kwargs=kwargs,
                    max_instances=max_instances,
                )
                if je is not None:
                    return je.id
                else:
                    return None
        else:
            logManager.error("Unknown type of job in add_dated_job")
            return None



from typing import Annotated
from fastapi import Depends
from aurori.users.dependencies import get_manadatory_user, get_optional_user

UserDep = Annotated[dict, Depends(get_manadatory_user)]
OptionalUserDep = Annotated[dict, Depends(get_optional_user)]



"""
The aurori project

Copyright (C) 2022  Marcus Drobisch,

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

__authors__ = ["Marcus Drobisch"]
__contact__ = "aurori@fabba.space"
__credits__ = []
__license__ = "AGPLv3+"

from aurori import db


class IntFlag(db.TypeDecorator):
    """
    Enables passing in a Python enum and storing the enum's *value* in the db.

    Usage example: message_type = db.Column(IntFlag(MessageTypes), default=MessageTypes.text)

    """
    impl = db.Integer

    def __init__(self, enumtype, *args, **kwargs):
        super(IntFlag, self).__init__(*args, **kwargs)
        self._enumtype = enumtype

    def process_bind_param(self, value, dialect):
        if isinstance(value, int):
            return value

        return value.value

    def process_result_value(self, value, dialect):
        return self._enumtype(value)



"""
The aurori project

Copyright (C) 2022  Marcus Drobisch,

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

__authors__ = ["Marcus Drobisch"]
__contact__ = "aurori@fabba.space"
__credits__ = []
__license__ = "AGPLv3+"

import argparse
import hashlib
import json
import re

line_splitter_flake8 = re.compile(
    '(.*):([0-9]+):([0-9]+):\s([A-Z][0-9]+)\s(.*)')
eslint_severity_splitter = re.compile('([a-zA-Z]*)\/?(.*)?')

parser = argparse.ArgumentParser()
parser.add_argument('filename_backend')
args = parser.parse_args()

exit_with_failure = False

with open(args.filename_backend) as flake_report, \
        open('code-quality-report.json', 'w') as quality_file:
    issues = []
    for line in flake_report:
        line_match = line_splitter_flake8.match(line)
        if line_match:
            tokens = line_match.groups()

            issue_code = tokens[3]
            issue_text = tokens[4]
            issue_hash = hashlib.md5(line.encode('utf-8')).hexdigest()

            if issue_code[0] in ['C']:
                issue_type = "critical"
            elif issue_code[0] in ['E', 'F']:
                issue_type = "major"
            elif issue_code[0] in ['W', 'N']:
                issue_type = "minor"
            else:
                issue_type = "critical"

            exit_with_failure = True
            i = {}
            i["description"] = "[{}] {}".format(issue_code, issue_text)
            i["fingerprint"] = issue_hash
            i["severity"] = issue_type
            i["location"] = {}
            i["location"]["path"] = tokens[0]
            i["location"]["lines"] = {"begin": int(tokens[1])}
            issues.append(i)

    # dump all the issues
    json.dump(issues, quality_file, indent=2)

if exit_with_failure is True:
    exit(0)



"""
The aurori project

Copyright (C) 2022  Marcus Drobisch,

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

__authors__ = ["Marcus Drobisch"]
__contact__ = "aurori@fabba.space"
__credits__ = []
__license__ = "AGPLv3+"

import json
from aurori import db
from sqlalchemy.ext import mutable
from sqlalchemy import Text, TypeDecorator

class JsonDict(TypeDecorator):
    """Enables JSON storage by encoding and decoding on the fly."""
    impl = Text

    def process_bind_param(self, value, dialect):
        if value is None:
            return '{}'
        else:
            return json.dumps(value)

    def process_result_value(self, value, dialect):
        if value is None:
            return {}
        else:
            return json.loads(value)


mutable.MutableDict.associate_with(JsonDict)



"""
The aurori project

Copyright (C) 2022  Marcus Drobisch,

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

__authors__ = ["Marcus Drobisch"]
__contact__ = "aurori@fabba.space"
__credits__ = []
__license__ = "AGPLv3+"

import logging
import sys
from fastapi import APIRouter, FastAPI
from os import environ

from sqlalchemy_utils import database_exists

from aurori.config import configure_app, config_manager
from aurori.version import version
from aurori.logs import logManager
from aurori.workspaces import workspaceManager
from aurori.users import userManager
from aurori.workspaces.permission import Permission
from aurori.jobs import jobManager
from aurori.database import db, SQLModelBase
from aurori.utils import unicode_string_to_bytes
from aurori.pages import menu_builder

__version__ = version


def create_app(config, fastapi_config=None, minimal=False, api_prefix="", title="Aurori App", version="0.1.0"):
    config_manager.init_manager(config, fastapi_config)
    app = FastAPI(title=title, version=version)

    db.init_app(config)

    from aurori import app_router
    app.include_router(app_router, tags=["Core"])

    from aurori.jobs import jobs_router
    app.include_router(jobs_router,prefix=api_prefix, tags=["Jobs"])

    from aurori.users import user_router, core_router, basemodels  # noqa: F401, F811
    app.include_router(core_router, tags=["Core"])
    app.include_router(user_router, prefix=api_prefix + "/user", tags=["App.User"])

    if not minimal:
        jobManager.init_manager(config_manager.config)
        menu_builder.init_builder(app, db, userManager, workspaceManager)

    workspaceManager.init_app(app, config_manager.config, minimal=minimal,api_prefix=api_prefix)
    if not minimal:
        userManager.init_manager(app, db, workspaceManager, config_manager.config)        


    database_is_new = not database_exists(db.url)
    if database_is_new == True:
        logManager.warning("No database found at", db.url)
        logManager.warning("Creating new empty database")
        SQLModelBase.metadata.create_all(bind=db.engine)

    return app


# declare app routes
app_router = APIRouter()
from aurori import routes



from typing import Annotated
from fastapi import Depends, HTTPException, status
from fastapi.security import OAuth2PasswordBearer
import jwt
from aurori.users import userManager
from aurori.config import config_manager
from .typemodels import Token, TokenData

oauth2_scheme = OAuth2PasswordBearer(tokenUrl="/token")

async def get_manadatory_user(token: Annotated[str, Depends(oauth2_scheme)]):
    credentials_exception = HTTPException(
        status_code=status.HTTP_401_UNAUTHORIZED,
        detail="Could not validate credentials",
        headers={"WWW-Authenticate": "Bearer"},
    )
    secret = config_manager.config['SECURITY']['secret_key']
    algorithm = config_manager.config['SECURITY']['jwt_algorithm']

    try:
        payload = jwt.decode(token, secret, algorithms=[algorithm])
        username: str = payload.get("sub")
        if username is None:
            raise credentials_exception
        token_data = TokenData(username=username)
    except jwt.PyJWTError:
        raise credentials_exception
    user = userManager.getUser(token_data.username) # get_user(fake_users_db, username=token_data.username)
    if user is None:
        raise credentials_exception
    return user


async def get_optional_user(token: Annotated[str, Depends(oauth2_scheme)]):
    credentials_exception = HTTPException(
        status_code=status.HTTP_401_UNAUTHORIZED,
        detail="Could not validate credentials",
        headers={"WWW-Authenticate": "Bearer"},
    )
    secret = config_manager.config['SECURITY']['secret_key']
    algorithm = config_manager.config['SECURITY']['jwt_algorithm']

    try:
        payload = jwt.decode(token, secret, algorithms=[algorithm])
        username: str = payload.get("sub")
        if username is None:
            raise credentials_exception
        token_data = TokenData(username=username)
    except jwt.PyJWTError:
        raise credentials_exception
    user = userManager.getUser(token_data.username) # get_user(fake_users_db, username=token_data.username)
    return user


"""
The aurori project

Copyright (C) 2022  Marcus Drobisch,

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

__authors__ = ["Marcus Drobisch"]
__contact__ = "aurori@fabba.space"
__credits__ = []
__license__ = "AGPLv3+"


class MenuBuilder(object):
    """ The MenuBuilder ...
    """
    def __init__(self, ):
        # preparation to instanciate
        self.actionsMap = {}

    def init_builder(self, app, db, userManager, workspaceManager):
        self.app = app
        self.db = db
        self.workspaceManager = workspaceManager
        self.userManager = userManager

    def make_menu_header(self, name):
        header = {}
        header['header'] = name
        return header

    def make_menu_group(self, name):
        group = {}
        group['title'] = name
        group['entries'] = []
        group['active'] = False
        group['disabled'] = False  # whether group can be collapsed
        return group

    def make_menu_divider(self, ):
        divider = {}
        divider['divider'] = True
        return divider

    def make_menu_entry(self, title, group, name, icon, path, external=False):
        entry = {}
        entry['title'] = title
        entry['group'] = group
        entry['name'] = name
        entry['icon'] = icon
        entry['href'] = path
        entry['external'] = external
        return entry

    def check_users_permissions(self, user, requirement, workspace):
        key = workspace.name + '.' + requirement.name
        for g in user.permission_groups:
            for p in g.permissions:
                if key == p.name:
                    return True
        return False

    def has_permission(self, user, page, workspace):
        if user is None:
            if page.requireLogin is False:
                return True
        else:
            if user.admin is True:
                return True
            if page.requireAdmin is False:
                if hasattr(page, 'requirePermission'):
                    if page.requirePermission is None:
                        return True
                    else:
                        return self.check_users_permissions(
                            user, page.requirePermission, workspace)
        return False

    def build_menu(self, user):
        menu_groups = {}
        for w in self.workspaceManager.workspaces:
            for key, p in w.pages.items():
                # is pages in a group
                if p.group is not None:
                    if str(p.group) in menu_groups:
                        rankSum, pageList = menu_groups[str(p.group)]

                        if self.has_permission(user, p, w):
                            pageList.append(p)

                        pageList.sort(key=lambda x: x.rank, reverse=True)
                        menu_groups[str(p.group)] = rankSum + p.rank, pageList
                    else:
                        if self.has_permission(user, p, w):
                            menu_groups[str(p.group)] = p.rank, [p]
                        else:
                            menu_groups[str(p.group)] = p.rank, []
                else:
                    if self.has_permission(user, p, w):
                        menu_groups['_' + str(p.name)] = p.rank, [p]

        sorted_menu_groups = sorted(menu_groups.items(),
                                    key=lambda x: x[1][0],
                                    reverse=True)

        menu = []
        for key, element in enumerate(sorted_menu_groups):
            group, (rank, pagelist) = element
            if len(pagelist) > 0:
                entry = self.make_menu_group(str(group))
                for p in pagelist:
                    entry['entries'].append(
                        self.make_menu_entry(title=p.title,
                                       group=p.group,
                                       name=p.name,
                                       icon=p.icon,
                                       path=p.route))
                menu.append(entry)

        return menu


'''
EXAMPLE of menu structure
[
  {header: 'aurori'},
  {
    title: 'Home',
    group: 'apps',
    icon: 'dashboard',
    name: 'Dashboard',
    href: 'dashboard'
  },
  {divider: true},
  {header: 'Templates&Links'},
  {
    title: 'Vue Material Admin',
    group: 'links',
    icon: 'touch_app',
    external: true,
    href: 'https://github.com/tookit/vue-material-admin'
  },
  {
    title: 'Vue Material Admin Demo',
    group: 'links',
    icon: 'touch_app',
    external: true,
    href: 'http://vma.isocked.com/#/dashboard'
  },
  {
    title: 'Vuetify',
    group: 'links',
    icon: 'touch_app',
    external: true,
    href: 'https://vuetifyjs.com/en/getting-started/quick-start'
  },
];

'''



"""
The aurori project

Copyright (C) 2022  Marcus Drobisch,

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

__authors__ = ["Marcus Drobisch"]
__contact__ = "aurori@fabba.space"
__credits__ = []
__license__ = "AGPLv3+"

import os
import pathlib
import pkgutil
import inspect
from aurori.workspaces.workspace import Workspace
from aurori.logs import logManager
import sys
import traceback
import sqlalchemy as sa


class WorkspaceManager(object):
    """ The WorkspaceManager holds all available workspaces and load them while creation.
    """
    def __init__(self, workspace_source_path):
        self.workspace_source_path = workspace_source_path
        self.app = None
        self.db = None
        self.config = None
        self.workspaces = None
        self.user_class = None
        self.api_prefix = None

    def init_app(self, app, config, minimal = False, api_prefix="api/v1"):
        self.app = app
        self.config = config
        self.workspaces = []
        self.seen_paths = []
        self.api_prefix = api_prefix
        logManager.info("")
        logManager.info(
            f'Discover workspaces in path : {self.workspace_source_path}')
        self.discover_workspaces(self.workspace_source_path)
        self.discover_models()
        if minimal is False:
            self.discover_permissions()
            self.create_permissions()
            self.discover_jobs()
            self.discover_pages()
            self.discover_apis()

        #    self.registerWorkspacePermissions()
        #    self.registerWorkspaceAPIs()

        logManager.info("")
        logManager.info("Workspaces and their components initialized")

    def get_workspace(self, name):
        for w in self.workspaces:
            if w.name is name:
                return w
        return None

    def reload_workspaces(self):
        """Reset the list of all plugins and initiate the walk over the main
        provided plugin package to load all available plugins
        """
        self.workspaces = []
        self.seen_paths = []
        logManager.info("")
        logManager.info(
            f'Discover workspaces in path : {self.workspace_source_path}')
        self.discover_workspaces(self.workspace_source_path)

    def discover_permissions(self):

        logManager.info("")
        logManager.info("Discover permissions:")

        for w in self.workspaces:
            logManager.info(f'  For workspace: "{w.name}"')

            # try to register permissions
            try:
                w.discover_permissions(self.workspace_source_path)
            except Exception as e:
                traceback.print_exc(file=sys.stdout)
                logManager.error(
                    f'  Workspace "{w.uri}" unable to discover permissions  ({str(type(e).__name__)}:{e})'
                )

    def discover_pages(self):
        logManager.info("")
        logManager.info("Discover pages:")

        for w in self.workspaces:
            logManager.info(f'  For workspace: "{w.name}"')
            workspace_config_section_name = pathlib.Path(
                w.path).resolve().name.upper()
            # try to register permissions

            try:
                w.discover_pages(self.workspace_source_path)
            except Exception as exception:
                traceback.print_exc(file=sys.stdout)
                logManager.error(
                    f'  Workspace "{w.name}" unable to discover pages  ({str(type(exception).__name__)}:{exception})'
                )

    def discover_jobs(self):
        logManager.info("")
        logManager.info("Discover jobs:")

        for w in self.workspaces:
            logManager.info(f'  For workspace: "{w.name}"')
            workspace_config_section_name = pathlib.Path(
                w.path).resolve().name.upper()

            if workspace_config_section_name in self.config and "disable_jobs" in self.config[
                    workspace_config_section_name] and self.config[
                        workspace_config_section_name]["disable_jobs"]:
                logManager.info(f'    Jobs for "{w.name}" are disabled by config')
            else:
                # try to register jobs
                try:
                    w.discover_jobs(self.workspace_source_path)
                except Exception as e:
                    traceback.print_exc(file=sys.stdout)
                    logManager.error(
                        f'    Workspace "{w.name}" unable to discover jobs  ({str(type(e).__name__)}:{e})'
                    )

    def discover_apis(self):
        logManager.info("")
        logManager.info("Discover apis:")

        for w in self.workspaces:
            logManager.info(f'  For workspace: "{w.name}"')
            workspace_config_section_name = pathlib.Path(
                w.path).resolve().name.upper()
            
            try:
                w.discover_apis(self.workspace_source_path)
            except Exception as exception:
                traceback.print_exc(file=sys.stdout)
                logManager.error(
                    f'  Workspace "{w.name}" unable to apis pages  ({str(type(exception).__name__)}:{exception})'
                )
                
    def discover_models(self):
        source = self.workspace_source_path
        try:
            imported_source = __import__(source, fromlist=['blah'])
        except Exception as exception:
            logManager.error(f"Unable to locate workspaces {str(exception)}")
            return
        all_current_paths = []

        # all_current_paths.append(imported_source.__path__._path)

        if isinstance(imported_source.__path__, str):
            all_current_paths.append(imported_source.__path__)
        else:
            all_current_paths.extend([x for x in imported_source.__path__])

        from aurori.users.basemodels import UserBase
        
        # remove duplicates
        all_current_paths = list(set(all_current_paths))

        for pkg_path in all_current_paths:
            # Walk through all sub directories
            child_pkgs = [
                p for p in os.listdir(pkg_path)
                if os.path.isdir(os.path.join(pkg_path, p))
            ]

            for child_pkg in child_pkgs:
                try:
                    # fixme: this was assigned to a later unused variable?
                    # imported_package = __import__(source + '.' + child_pkg + '.models', fromlist=['blah'])
                    imported_package = __import__(source + '.' + child_pkg + '.models',
                               fromlist=['blah'])
                    for _, modelpackagename, ispkg in pkgutil.iter_modules(
                            imported_package.__path__,
                            imported_package.__name__ + '.'):
                        workspace_module = __import__(modelpackagename,
                                                      fromlist=['blah'])
                        clsmembers = inspect.getmembers(
                            workspace_module, inspect.isclass)

                        for (_, c) in clsmembers:     
                            if issubclass(c, UserBase) & (c is not UserBase):
                                if self.user_class is not None and self.user_class != c:
                                    raise Exception("Multiple User models found. For a aurori app only one is allowed.")
                                self.user_class = c
                except ModuleNotFoundError:
                    modelmodule = source + '.' + child_pkg + '.models'
                    logManager.info(f'No model found for {modelmodule}')
        
        if self.user_class is None:
            logManager.warning(f'No User model derived from aurori.users.basemodels.UserBase found.')
            logManager.warning(f'A aurori app needs a user model to work properly.')
            logManager.warning(f'Please define a derived model named "User" in your workspaces, if needed.')
            logManager.warning(f'The default model will be used, please reference: aurori.users.defaultmodels.User ')
            from aurori.users.defaultmodels import User

    def discover_workspaces(self, source):
        """Recursively walk the supplied package to retrieve all plugins (workspaces)
        """

        try:
            imported_source = __import__(source, fromlist=['blah'])
        except Exception as e:
            logManager.error(f"Unable to locate workspaces {str(e)}")
            return

        all_current_paths = []

        # all_current_paths.append(imported_source.__path__._path)

        if isinstance(imported_source.__path__, str):
            all_current_paths.append(imported_source.__path__)
        else:
            all_current_paths.extend([x for x in imported_source.__path__])

        # remove duplicates
        all_current_paths = list(set(all_current_paths))

        for pkg_path in all_current_paths:
            # Walk through all sub directories
            child_pkgs = [
                p for p in os.listdir(pkg_path)
                if os.path.isdir(os.path.join(pkg_path, p))
            ]

            # Every sub directory contains one workspace
            for child_pkg in child_pkgs:
                imported_package = __import__(source + '.' + child_pkg,
                                              fromlist=['blah'])
                
                clsmembers = inspect.getmembers(imported_package)
              
                uri = ""
                if hasattr(imported_package,"uri"):
                    uri = imported_package.uri
                else:
                    uri = str(child_pkg).lower()

                logger = None
                if hasattr(imported_package,"logger"):
                    logger = imported_package.logger


                workspaceInstance = Workspace(self.app, self.db, str(child_pkg), uri)

                if hasattr(imported_package,"description"):
                    workspaceInstance.description = imported_package.description

                if hasattr(imported_package,"disabled"):
                    workspaceInstance.disabled = imported_package.disabled

                workspaceInstance.path = os.path.dirname(
                    imported_package.__file__)
                workspace_config_section_name = pathlib.Path(
                    imported_package.__file__).parent.resolve(
                    ).name.upper()
                logManager.info(
                    f'  Discovered workspace "{workspaceInstance.name}" as "{workspaceInstance.uri}" from "{source + "." + child_pkg}"')

                if (workspace_config_section_name
                        in self.config) and (
                            "disabled" in self.config[
                                workspace_config_section_name]
                        ) and (self.config[
                            workspace_config_section_name]
                                ["disabled"]):
                    logManager.info(
                        '  The workspace is disabled by config and wont show up'
                    )
                else:
                    if workspaceInstance.disabled is True:
                        logManager.info(
                            '  The workspace is disabled by module definition and wont show up'
                        )
                    else:
                        logManager.register_logger(logger, workspaceInstance.name)
                        self.workspaces.append(
                            workspaceInstance)

    def create_permissions(self):
        """ Run createPermissions for all workspaces and store permissions
        """
        all_permissions = {}
        for ws in self.workspaces:

            workspace_permissions = ws.permissions
            if workspace_permissions is not None:
                all_permissions = {**all_permissions, **workspace_permissions}

        # delete orphaned permissions for security reasons
        from .models import Permission
        from aurori import db
        engine = db.engine
        table_exists = sa.inspect(engine).has_table(Permission.__tablename__)
        if table_exists:
            with db.get_session() as db_session:
                db_permissions = db_session.query(Permission).all()
                for permission in db_permissions:
                    if permission.name not in all_permissions:
                        logManager.warning(f'  Delete orphaned {permission}')
                        db_session.delete(permission)
                        db_session.commit()



from contextlib import contextmanager
from sqlalchemy import create_engine
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import sessionmaker, scoped_session

class SQLAlchemy(object):
    """ 
    """

    def __init__(self, ):
        self.engine = None
        self.url = ""
        self.sessionmaker = None

    def init_app(self, config):
        self.url = "sqlite:///./sqlite3.db"

        self.engine = create_engine(
            self.url, connect_args={"check_same_thread": False}
        )
        self.sessionmaker = scoped_session(sessionmaker(autocommit=False, autoflush=False, bind=self.engine))

    @contextmanager
    def get_session(self):
        db_session = (self.sessionmaker)()
        try:
            yield db_session
        finally:
            db_session.close()

db = SQLAlchemy()

SQLModelBase = declarative_base()



"""
The aurori project

Copyright (C) 2022  Marcus Drobisch,

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

__authors__ = ["Marcus Drobisch"]
__contact__ = "aurori@fabba.space"
__credits__ = []
__license__ = "AGPLv3+"

import json
import jwt
from typing import Annotated
from datetime import datetime, timedelta
from aurori.users import user_router, core_router
from fastapi import Depends, HTTPException, status
from fastapi.security import OAuth2PasswordBearer, OAuth2PasswordRequestForm
from pydantic import BaseModel
from .typemodels import Token, TokenData
from aurori.users import userManager


def authenticate_user(username: str, password: str):
    user = userManager.getUser(username)

    if user is not None:
        if userManager.checkUserPassword(user, password):
            return user
        else:
            return False
    else:
        return False


from aurori.config import config_manager


def create_access_token(data: dict, expires_delta: timedelta | None = None):
    to_encode = data.copy()
    if expires_delta:
        expire = datetime.utcnow() + expires_delta
    else:
        expire = datetime.utcnow() + timedelta(minutes=15)
    to_encode.update({"exp": expire})
    secret = config_manager.config["SECURITY"]["secret_key"]
    algorithm = config_manager.config["SECURITY"]["jwt_algorithm"]
    encoded_jwt = jwt.encode(to_encode, secret, algorithm=algorithm)
    return encoded_jwt


@core_router.post("/token", include_in_schema=True)
@user_router.post("/login", response_model=Token)
async def login_for_access_token(
    form_data: Annotated[OAuth2PasswordRequestForm, Depends()]
):
    user = authenticate_user(form_data.username, form_data.password)
    if not user:
        raise HTTPException(
            status_code=status.HTTP_401_UNAUTHORIZED,
            detail="Incorrect username or password",
            headers={"WWW-Authenticate": "Bearer"},
        )

    jwt_expires_in_minutes = config_manager.config["SECURITY"]["jwt_expires_in_minutes"]
    access_token_expires = timedelta(minutes=jwt_expires_in_minutes)
    access_token = create_access_token(
        data={"sub": user.email}, expires_delta=access_token_expires
    )
    print(access_token)
    return {"access_token": access_token, "token_type": "bearer"}
